import os
import os.path
import pathlib
from dataclasses import dataclass, field
from typing import Any, Optional, Self, Tuple, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class FileStep(Step):
    filename: str
    contents: str

    @staticmethod
    def get_details(step: Any) -> Tuple[str, str]:
        if "filename" not in step:
            raise ValueError('Missing "filename" field in file step.')

        if step["filename"] is None or step["filename"].strip() == "":
            raise ValueError('Empty "filename" field in file step.')

        filename = step["filename"]
        contents = step.get("contents", "") or ""
        return filename, contents


@dataclass
class NewFileStep(FileStep):
    step_type: StepType = field(init=False, default=StepType.NEW_FILE)

    def execute(self, repo: Repo) -> None:
        rw_dir = repo.working_dir
        filepath = os.path.join(rw_dir, self.filename)
        filepath_dir_only = os.path.dirname(filepath)
        pathlib.Path(filepath_dir_only).mkdir(parents=True, exist_ok=True)
        with open(filepath, "w+") as fs:
            fs.write(self.contents)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        filename, contents = FileStep.get_details(step)
        return cls(
            name=name,
            description=description,
            id=id,
            filename=filename,
            contents=contents,
        )


@dataclass
class EditFileStep(FileStep):
    step_type: StepType = field(init=False, default=StepType.EDIT_FILE)

    def execute(self, repo: Repo) -> None:
        rw_dir = repo.working_dir
        filepath = os.path.join(rw_dir, self.filename)
        if not os.path.isfile(filepath):
            raise ValueError("Invalid filename for editing")
        with open(filepath, "w") as fs:
            fs.write(self.contents)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        filename, contents = FileStep.get_details(step)
        return cls(
            name=name,
            description=description,
            id=id,
            filename=filename,
            contents=contents,
        )


@dataclass
class DeleteFileStep(FileStep):
    step_type: StepType = field(init=False, default=StepType.DELETE_FILE)

    def execute(self, repo: Repo) -> None:
        rw_dir = repo.working_dir
        filepath = os.path.join(rw_dir, self.filename)
        if not os.path.isfile(filepath):
            raise ValueError("Invalid filename for deleting")
        os.remove(filepath)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        filename, contents = FileStep.get_details(step)
        return cls(
            name=name,
            description=description,
            id=id,
            filename=filename,
            contents=contents,
        )


@dataclass
class AppendFileStep(FileStep):
    step_type: StepType = field(init=False, default=StepType.APPEND_FILE)

    def execute(self, repo: Repo) -> None:
        rw_dir = repo.working_dir
        filepath = os.path.join(rw_dir, self.filename)
        if not os.path.isfile(filepath):
            raise ValueError("Invalid filename for appending")
        with open(filepath, "a") as fs:
            fs.write(self.contents)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        filename, contents = FileStep.get_details(step)
        return cls(
            name=name,
            description=description,
            id=id,
            filename=filename,
            contents=contents,
        )



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class CommitOptions(TypedDict, total=False):
    all: bool
    reuse_message: str
    message: str
    allow_empty: bool


COMMIT_SPEC = (
    CommandSpec()
    .flag("all", "-a", default=False)
    .opt("reuse_message", "--reuse-message")
    .opt("message", "-m")
    .flag("allow_empty", "--allow-empty", default=False)
)



from src.repo_smith.initialize_repo import initialize_repo


def test_clone_from():
    repo_initializer = initialize_repo("tests/specs/clone_from.yml")
    with repo_initializer.initialize() as r:
        commits = list(r.iter_commits("main"))
        # Should be more than 1 (the empty commit we made)
        assert len(commits) > 1



import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_step_missing_type():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/step_missing_type.yml")



import os
import textwrap
from io import TextIOWrapper
from typing import Optional

from repo_smith.helpers.helper import Helper
from repo_smith.types import FilePath


class FilesHelper(Helper):
    def __init__(self, verbose: bool) -> None:
        super().__init__(verbose)

    def create_or_update(
        self, filepath: FilePath, contents: Optional[str] = None
    ) -> None:
        """Creates or updates a file with the given content."""
        dirname = os.path.dirname(filepath)
        if dirname != "":
            os.makedirs(dirname, exist_ok=True)

        if contents is None:
            open(filepath, "a").close()
        else:
            with open(filepath, "w") as file:
                self.__write_to_file__(file, contents)

    def append(self, filepath: FilePath, contents: str) -> None:
        """Appends contents to a given file."""
        with open(filepath, "a") as file:
            self.__write_to_file__(file, contents)

    def delete(self, filepath: FilePath) -> None:
        """Deletes a given file."""
        os.remove(filepath)

    def __write_to_file__(self, file: TextIOWrapper, contents: str) -> None:
        file.write(textwrap.dedent(contents).lstrip())



from typing import Dict, List

from repo_smith.command_result import CommandResult, run


class Helper:
    def __init__(self, verbose: bool) -> None:
        self.verbose = verbose

    def run(
        self,
        command: List[str],
        env: Dict[str, str] = {},
        exit_on_error: bool = False,
    ) -> CommandResult:
        return run(command, self.verbose, env, exit_on_error)




import pytest

from repo_smith.steps.add_step import AddStep


def test_add_step_parse_missing_files():
    with pytest.raises(ValueError, match='Missing "files" field in add step.'):
        AddStep.parse("a", "d", "id", {})


def test_add_step_parse_empty_files():
    with pytest.raises(ValueError, match='Empty "files" list in add step.'):
        AddStep.parse("a", "d", "id", {"files": []})


def test_add_step_parse():
    step = AddStep.parse("a", "d", "id", {"files": ["hello.txt"]})
    assert isinstance(step, AddStep)
    assert step.name == "a"
    assert step.description == "d"
    assert step.id == "id"
    assert step.files == ["hello.txt"]



from subprocess import CompletedProcess
from unittest.mock import patch

import pytest
from repo_smith.command_result import CommandResult
from repo_smith.helpers.git_helper.git_helper import GitHelper
from repo_smith.helpers.helper import Helper


def test_tag_with_annotate():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(["git", "tag", "v0.1.0", "-a"], returncode=0)
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.tag("v0.1.0", annotate=True)
        mock_helper.assert_called_with(["git", "tag", "v0.1.0", "-a"])


def test_tag_with_force():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(["git", "tag", "v0.1.0", "-f"], returncode=0)
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.tag("v0.1.0", force=True)
        mock_helper.assert_called_with(["git", "tag", "v0.1.0", "-f"])


def test_tag_with_message():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "tag", "v0.1.0", "-m", "This is a message."], returncode=0
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.tag("v0.1.0", message="This is a message.")
        mock_helper.assert_called_with(
            ["git", "tag", "v0.1.0", "-m", "This is a message."]
        )


def test_tag_with_multiple_flags():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "tag", "v0.1.0", "-m", "This is a message.", "-f", "-a"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.tag("v0.1.0", message="This is a message.", force=True, annotate=True)
        mock_helper.assert_called_with(
            ["git", "tag", "v0.1.0", "-f", "-m", "This is a message.", "-a"]
        )


def test_add_all():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "add", "-a"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.add(all=True)
        mock_helper.assert_called_with(["git", "add", "-a"])


def test_commit_all():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "commit", "-a", "-m", "Test commit"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.commit(all=True, message="Test commit")
        mock_helper.assert_called_with(["git", "commit", "-a", "-m", "Test commit"])


def test_commit_with_pathspec():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "commit", "-a", "-m", "Test commit", "file*"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.commit(pathspec="file*", all=True, message="Test commit")
        mock_helper.assert_called_with(
            ["git", "commit", "-a", "-m", "Test commit", "file*"]
        )


def test_checkout_with_branch_and_start_point():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "checkout", "-b", "test", "HEAD"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.checkout(
            branch_name="test",
            start_point="HEAD",
            branch=True,
        )
        mock_helper.assert_called_with(["git", "checkout", "-b", "test", "HEAD"])


def test_checkout_with_files():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "checkout", "test", "--", "filea.txt", "fileb.txt"],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.checkout(
            branch_name="test",
            paths=["filea.txt", "fileb.txt"],
        )
        mock_helper.assert_called_with(
            ["git", "checkout", "test", "--", "filea.txt", "fileb.txt"]
        )


def test_restore_staged_worktree():
    with patch.object(
        Helper,
        "run",
        return_value=CommandResult(
            CompletedProcess(
                ["git", "restore", "-S", "-W", "."],
                returncode=0,
            )
        ),
    ) as mock_helper:
        gh = GitHelper(False)
        gh.restore(pathspec=".", staged=True, worktree=True)
        mock_helper.assert_called_with(["git", "restore", "-W", "-S", "."])


def test_restore_ours():
    with (
        patch.object(
            Helper,
            "run",
            return_value=CommandResult(
                CompletedProcess(
                    ["git", "restore", "-S", "-W", "."],
                    returncode=0,
                )
            ),
        ),
        pytest.raises(ValueError),
    ):
        gh = GitHelper(False)
        gh.restore(pathspec=".", ours=True, source=".", staged=True, worktree=True)



import subprocess
from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class BashStep(Step):
    body: str

    step_type: StepType = field(init=False, default=StepType.BASH)

    def execute(self, repo: Repo) -> None:
        subprocess.check_call(
            self.body.strip(), shell=True, executable="/bin/bash", cwd=repo.working_dir
        )

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "runs" not in step:
            raise ValueError('Missing "runs" field in bash step.')

        if step["runs"] is None or step["runs"].strip() == "":
            raise ValueError('Empty "runs" field in bash step.')

        return cls(
            name=name,
            description=description,
            id=id,
            body=step["runs"],
        )



import os
import shutil
from pathlib import Path
from typing import Generator

import pytest
from git import Repo

DUMMY_PATH = Path("tests/dummy")
REMOTE_REPO_PATH = DUMMY_PATH / "remote_repo"


@pytest.fixture
def remote_repo() -> Generator[Repo, None, None]:
    os.makedirs(REMOTE_REPO_PATH)
    repo = Repo.init(REMOTE_REPO_PATH, initial_branch="main")

    yield repo

    shutil.rmtree(REMOTE_REPO_PATH, ignore_errors=True)
    shutil.rmtree(DUMMY_PATH, ignore_errors=True)



import pytest

from repo_smith.steps.merge_step import MergeStep


def test_merge_step_parse_missing_branch_name():
    with pytest.raises(ValueError, match='Missing "branch-name" field in merge step.'):
        MergeStep.parse("n", "d", "id", {})


def test_merge_step_parse_empty_branch_name():
    with pytest.raises(ValueError, match='Empty "branch-name" field in merge step.'):
        MergeStep.parse("n", "d", "id", {"branch-name": ""})


MERGE_STEP_CONFIGURATION = {
    "no-ff not set, squash not set": {"branch-name": "test"},
    "no-ff not set, squash set": {"branch-name": "test", "squash": True},
    "no-ff set, squash not set": {"branch-name": "test", "no-ff": True},
    "no-ff set, squash set": {"branch-name": "test", "squash": True, "no-ff": True},
}


@pytest.mark.parametrize("config_name, config", MERGE_STEP_CONFIGURATION.items())
def test_merge_step_parse(config_name, config):
    step = MergeStep.parse("n", "d", "id", config)
    assert isinstance(step, MergeStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.branch_name == "test"
    assert step.no_fast_forward == config.get("no-ff", False)
    assert step.squash == config.get("squash", False)



import re
from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class TagStep(Step):
    tag_name: str
    tag_message: Optional[str]

    step_type: StepType = field(init=False, default=StepType.TAG)

    def execute(self, repo: Repo) -> None:
        repo.create_tag(self.tag_name, message=self.tag_message)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "tag-name" not in step:
            raise ValueError('Missing "tag-name" field in tag step.')

        if step["tag-name"] is None or step["tag-name"].strip() == "":
            raise ValueError('Empty "tag-name" field in tag step.')

        tag_name_regex = "^[0-9a-zA-Z-_.]*$"
        if re.search(tag_name_regex, step["tag-name"]) is None:
            raise ValueError(
                'Field "tag-name" can only contain alphanumeric characters, _, -, .'
            )

        return cls(
            name=name,
            description=description,
            id=id,
            tag_name=step["tag-name"],
            tag_message=step.get("tag-message"),
        )



import os

import pytest
from git import Repo

from src.repo_smith.initialize_repo import initialize_repo


def test_delete_file_step_missing_filename() -> None:
    with pytest.raises(Exception):
        initialize_repo("tests/specs/delete_file_step/delete_file_missing_filename.yml")


def test_delete_file_step_unknown_file() -> None:
    with pytest.raises(Exception):
        repo_initializer = initialize_repo(
            "tests/specs/delete_file_step/delete_file_no_file.yml"
        )
        with repo_initializer.initialize():
            pass


def test_delete_file_step() -> None:
    def pre_hook(r: Repo) -> None:
        filepath = os.path.join(r.working_dir, "file.txt")
        assert os.path.isfile(filepath)

    repo_initializer = initialize_repo("tests/specs/delete_file_step/delete_file.yml")
    repo_initializer.add_pre_hook("delete-file", pre_hook)
    with repo_initializer.initialize() as r:
        filepath = os.path.join(r.working_dir, "file.txt")
        assert not os.path.isfile(filepath)



import os
import shutil
import tempfile
from contextlib import contextmanager
from typing import Any, Callable, Dict, Iterator, Optional, Set, TypeAlias

import yaml
from git import Repo

import repo_smith.steps.tag_step
from repo_smith.clone_from import CloneFrom
from repo_smith.spec import Spec
from repo_smith.steps.dispatcher import Dispatcher

Hook: TypeAlias = Callable[[Repo], None]


class RepoInitializer:
    def __init__(self, spec_data: Any) -> None:
        self.__spec_data = spec_data
        self.__pre_hooks: Dict[str, Hook] = {}
        self.__post_hooks: Dict[str, Hook] = {}

        self.__spec = self.__parse_spec(self.__spec_data)
        self.__validate_spec(self.__spec)
        self.__step_ids = self.__get_all_ids(self.__spec)

    @contextmanager
    def initialize(self, existing_path: Optional[str] = None) -> Iterator[Repo]:
        tmp_dir = tempfile.mkdtemp() if existing_path is None else existing_path
        repo: Optional[Repo] = None
        try:
            if self.__spec.clone_from is not None:
                repo = Repo.clone_from(self.__spec.clone_from.repo_url, tmp_dir)
            else:
                repo = Repo.init(tmp_dir, initial_branch="main")

            for step in self.__spec.steps:
                if step.id in self.__pre_hooks:
                    self.__pre_hooks[step.id](repo)

                step.execute(repo=repo)

                if step.id in self.__post_hooks:
                    self.__post_hooks[step.id](repo)
            yield repo
        finally:
            if repo is not None:
                repo.git.clear_cache()
                shutil.rmtree(tmp_dir)

    def add_pre_hook(self, id: str, hook: Hook) -> None:
        if id not in self.__step_ids:
            ids = "\n".join([f"- {id}" for id in self.__step_ids])
            raise ValueError(
                f"ID {id} not found in spec's steps. Available IDs:\n{ids}"
            )

        if id in self.__pre_hooks:
            raise ValueError(
                f"ID {id} already has a pre-hook set. Did you mean to add a post_hook instead?"
            )

        self.__pre_hooks[id] = hook

    def add_post_hook(self, id: str, hook: Hook) -> None:
        if id not in self.__step_ids:
            ids = "\n".join([f"- {id}" for id in self.__step_ids])
            raise ValueError(
                f"ID {id} not found in spec's steps. Available IDs:\n{ids}"
            )

        if id in self.__post_hooks:
            raise ValueError(
                f"ID {id} already has a post-hook set. Did you mean to add a pre_hook instead?"
            )

        self.__post_hooks[id] = hook

    def __validate_spec(self, spec: Spec) -> None:
        ids: Set[str] = set()
        tags: Set[str] = set()
        for step in spec.steps:
            if step.id is not None:
                if step.id in ids:
                    raise ValueError(
                        f"ID {step.id} is duplicated from a previous step. All IDs should be unique."
                    )
                ids.add(step.id)

            if isinstance(step, repo_smith.steps.tag_step.TagStep):
                if step.tag_name in tags:
                    raise ValueError(
                        f"Tag {step.tag_name} is already in use by a previous step. All tag names should be unique."
                    )
                tags.add(step.tag_name)

    def __get_all_ids(self, spec: Spec) -> Set[str]:
        ids = set()
        for step in spec.steps:
            if step.id is not None:
                ids.add(step.id)
        return ids

    def __parse_spec(self, spec: Any) -> Spec:
        steps = []

        for step in spec.get("initialization", {}).get("steps", []) or []:
            steps.append(Dispatcher.dispatch(step))

        clone_from = None
        if spec.get("initialization", {}).get("clone-from", None) is not None:
            clone_from = CloneFrom(
                repo_url=spec.get("initialization", {}).get("clone-from", "")
            )

        return Spec(
            name=spec.get("name", "") or "",
            description=spec.get("description", "") or "",
            steps=steps,
            clone_from=clone_from,
        )


def initialize_repo(spec_path: str) -> RepoInitializer:
    if not os.path.isfile(spec_path):
        raise ValueError("Invalid spec_path provided, not found.")

    with open(spec_path, "rb") as spec_file:
        try:
            spec_data = yaml.safe_load(spec_file)
            if spec_data is None:
                raise ValueError("Incomplete spec file.")
            return RepoInitializer(spec_data)
        except Exception as e:
            raise e



from repo_smith.initialize_repo import initialize_repo


def test_remote_step_valid():
    ir = initialize_repo("tests/specs/remote_step/remote_step_valid.yml")
    with ir.initialize() as r:
        assert len(r.remotes) == 1
        assert r.remotes[0].name == "upstream"
        assert r.remotes[0].url == "https://github.com/git-mastery/repo-smith.git"



from typing import Literal, TypedDict, Union

from repo_smith.helpers.command_spec import CommandSpec


class RestoreOptions(TypedDict, total=False):
    source: str

    worktree: bool
    staged: bool

    ours: bool
    theirs: bool

    merge: bool
    conflict: Union[Literal["merge"], Literal["diff3"], Literal["zdiff3"]]

    ignore_unmerged: bool
    ignore_skip_worktree_bits: bool

    recurse_submodules: bool
    no_recurse_submodules: bool

    overlay: bool
    no_overlay: bool


RESTORE_SPEC = (
    CommandSpec()
    .opt("source", "--source")
    .flag("worktree", "-W", default=False)
    .flag("staged", "-S", default=False)
    .flag("ours", "--ours", default=False)
    .flag("theirs", "--theirs", default=False)
    .flag("merge", "--merge", default=False)
    .opt("conflict", "--conflict")
    .flag("ignore_unmerged", "--ignore-unmerged", default=False)
    .flag("ignore_skip_worktree_bits", "--ignore-skip-worktree-bits", default=False)
    .flag("recurse_submodules", "--recurse-submodules", default=False)
    .flag("no_recurse_submodules", "--no-recurse-submodules", default=False)
    .flag("overlay", "--overlay", default=False)
    .flag("no_overlay", "--no-overlay", default=False)
)



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class MergeStep(Step):
    branch_name: str
    no_fast_forward: bool
    squash: bool

    step_type: StepType = field(init=False, default=StepType.MERGE)

    def execute(self, repo: Repo) -> None:
        # TODO: Maybe handle merge conflicts as they happen
        merge_args = [self.branch_name, "--no-edit"]

        if self.squash:
            merge_args.append("--squash")
        elif self.no_fast_forward:
            merge_args.append("--no-ff")

        repo.git.merge(*merge_args)

        if self.squash:
            repo.git.commit("-m", f"Squash merge branch '{self.branch_name}'")

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "branch-name" not in step:
            raise ValueError('Missing "branch-name" field in merge step.')

        if step["branch-name"] is None or step["branch-name"].strip() == "":
            raise ValueError('Empty "branch-name" field in merge step.')

        return cls(
            name=name,
            description=description,
            id=id,
            branch_name=step.get("branch-name"),
            no_fast_forward=step.get("no-ff", False),
            squash=step.get("squash", False),
        )



import pytest

from repo_smith.steps.commit_step import CommitStep


def test_commit_step_parse_missing_message():
    with pytest.raises(ValueError, match='Missing "message" field in commit step.'):
        CommitStep.parse("n", "d", "id", {})


def test_commit_step_parse_empty_message():
    with pytest.raises(ValueError, match='Empty "message" field in commit step.'):
        CommitStep.parse("n", "d", "id", {"message": ""})


def test_commit_step_parse_missing_empty():
    step = CommitStep.parse("n", "d", "id", {"message": "Test"})
    assert isinstance(step, CommitStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.message == "Test"
    assert not step.empty


def test_commit_step_parse_with_empty():
    step = CommitStep.parse("n", "d", "id", {"message": "Test", "empty": True})
    assert isinstance(step, CommitStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.message == "Test"
    assert step.empty



from typing import Any, Type

from repo_smith.steps.add_step import AddStep
from repo_smith.steps.bash_step import BashStep
from repo_smith.steps.branch_delete_step import BranchDeleteStep
from repo_smith.steps.branch_rename_step import BranchRenameStep
from repo_smith.steps.branch_step import BranchStep
from repo_smith.steps.checkout_step import CheckoutStep
from repo_smith.steps.commit_step import CommitStep
from repo_smith.steps.fetch_step import FetchStep
from repo_smith.steps.file_step import (
    AppendFileStep,
    DeleteFileStep,
    EditFileStep,
    NewFileStep,
)
from repo_smith.steps.merge_step import MergeStep
from repo_smith.steps.remote_step import RemoteStep
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType
from repo_smith.steps.tag_step import TagStep


class Dispatcher:
    @staticmethod
    def dispatch(step: Any) -> Step:
        if "type" not in step:
            raise ValueError('Missing "type" field in step.')

        name = step.get("name")
        description = step.get("description")
        step_type = StepType.from_value(step["type"])
        id = step.get("id")
        retrieved_step_type = Dispatcher.__get_type(step_type)
        return retrieved_step_type.parse(name, description, id, step)

    @staticmethod
    def __get_type(step_type: StepType) -> Type[Step]:
        match step_type:
            case StepType.COMMIT:
                return CommitStep
            case StepType.ADD:
                return AddStep
            case StepType.TAG:
                return TagStep
            case StepType.BASH:
                return BashStep
            case StepType.BRANCH:
                return BranchStep
            case StepType.BRANCH_RENAME:
                return BranchRenameStep
            case StepType.BRANCH_DELETE:
                return BranchDeleteStep
            case StepType.CHECKOUT:
                return CheckoutStep
            case StepType.MERGE:
                return MergeStep
            case StepType.REMOTE:
                return RemoteStep
            case StepType.FETCH:
                return FetchStep
            case StepType.NEW_FILE:
                return NewFileStep
            case StepType.EDIT_FILE:
                return EditFileStep
            case StepType.DELETE_FILE:
                return DeleteFileStep
            case StepType.APPEND_FILE:
                return AppendFileStep



import pytest

from repo_smith.steps.branch_delete_step import BranchDeleteStep


def test_branch_delete_step_parse_missing_branch_name():
    with pytest.raises(
        ValueError, match='Missing "branch-name" field in branch-delete step.'
    ):
        BranchDeleteStep.parse("n", "d", "id", {})


def test_branch_delete_step_parse_empty_branch_name():
    with pytest.raises(
        ValueError, match='Empty "branch-name" field in branch-delete step.'
    ):
        BranchDeleteStep.parse("n", "d", "id", {"branch-name": ""})


def test_branch_delete_step_parse():
    step = BranchDeleteStep.parse("n", "d", "id", {"branch-name": "test"})
    assert isinstance(step, BranchDeleteStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.branch_name == "test"




import pytest

from repo_smith.steps.step_type import StepType


def test_step_type_uncovered_type():
    with pytest.raises(ValueError):
        StepType.from_value("this should not be implemented")



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class CheckoutOptions(TypedDict, total=False):
    branch: bool


CHECKOUT_SPEC = CommandSpec().flag("branch", "-b", default=False)



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class RemoteAddOptions(TypedDict, total=False):
    fetch: bool
    tags: bool
    no_tags: bool


REMOTE_ADD_SPEC = (
    CommandSpec()
    .flag("fetch", "-f", default=False)
    .flag("tags", "--tags", default=False)
    .flag("no_tags", "--no-tags", default=False)
)



import pytest

from repo_smith.steps.branch_delete_step import BranchDeleteStep
from repo_smith.steps.branch_rename_step import BranchRenameStep


def test_branch_rename_step_parse_missing_branch_name():
    with pytest.raises(
        ValueError, match='Missing "branch-name" field in branch-rename step.'
    ):
        BranchRenameStep.parse("n", "d", "id", {})


def test_branch_rename_step_parse_empty_branch_name():
    with pytest.raises(
        ValueError, match='Empty "branch-name" field in branch-rename step.'
    ):
        BranchRenameStep.parse("n", "d", "id", {"branch-name": ""})


def test_branch_rename_step_parse_missing_new_name():
    with pytest.raises(
        ValueError, match='Missing "new-name" field in branch-rename step.'
    ):
        BranchRenameStep.parse("n", "d", "id", {"branch-name": "test"})


def test_branch_rename_step_parse_empty_new_name():
    with pytest.raises(
        ValueError, match='Empty "new-name" field in branch-rename step.'
    ):
        BranchRenameStep.parse("n", "d", "id", {"branch-name": "test", "new-name": ""})


def test_branch_rename_step_parse():
    step = BranchRenameStep.parse(
        "n", "d", "id", {"branch-name": "test", "new-name": "other"}
    )
    assert isinstance(step, BranchRenameStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.original_branch_name == "test"
    assert step.target_branch_name == "other"



from git import Repo
import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_commit_step_missing_message():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/commit_step/commit_step_missing_message.yml")


def test_commit_step():
    def pre_hook(r: Repo) -> None:
        with pytest.raises(Exception):
            # Ensure there is 0 commits
            # Must cast to list since the Iterator alone will not raise an error
            list(r.iter_commits("main"))

    repo_initializer = initialize_repo("tests/specs/commit_step/commit.yml")
    repo_initializer.add_pre_hook("commit", pre_hook)
    with repo_initializer.initialize() as r:
        commits = list(r.iter_commits("main"))
        assert len(commits) == 1
        commit = commits[0]
        assert len(commit.stats.files) == 1
        assert "file.txt" in commit.stats.files


def test_commit_step_empty_commit():
    def pre_hook(r: Repo) -> None:
        with pytest.raises(Exception):
            # Ensure there is 0 commits
            # Must cast to list since the Iterator alone will not raise an error
            list(r.iter_commits("main"))

    repo_initializer = initialize_repo("tests/specs/commit_step/commit_empty.yml")
    repo_initializer.add_pre_hook("commit", pre_hook)
    with repo_initializer.initialize() as r:
        commits = list(r.iter_commits("main"))
        assert len(commits) == 1
        commit = commits[0]
        assert len(commit.stats.files) == 0



import os

from git import Repo
import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_branch_step_missing_branch_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/branch_step/branch_step_missing_branch_name.yml")


def test_branch_step_empty_branch_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/branch_step/branch_step_empty_branch_name.yml")


def test_branch_step():
    repo_initializer = initialize_repo("tests/specs/branch_step/branch_step.yml")
    with repo_initializer.initialize() as r:
        assert len(r.branches) == 2
        assert "test" in r.heads



from dataclasses import dataclass
from typing import List, Optional

from repo_smith.clone_from import CloneFrom
from repo_smith.steps.step import Step


@dataclass
class Spec:
    name: str
    description: Optional[str]
    steps: List[Step]
    clone_from: Optional[CloneFrom]



import os
import subprocess

import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_bash_step_missing_runs():
    with pytest.raises(ValueError, match='Missing "runs" field in bash step.'):
        initialize_repo("tests/specs/bash_step/bash_step_missing_runs.yml")


@pytest.mark.skip(
    reason="an actual empty field is not parsed so we can safely ignore this"
)
def test_bash_step_empty_runs():
    with pytest.raises(ValueError, match='Empty "runs" field in bash step.'):
        initialize_repo("tests/specs/bash_step/bash_step_missing_runs.yml")


def test_bash_step_invalid_runs():
    with pytest.raises(subprocess.CalledProcessError):
        repo_initializer = initialize_repo(
            "tests/specs/bash_step/bash_step_invalid_runs.yml"
        )
        with repo_initializer.initialize():
            pass


def test_bash_step():
    repo_initializer = initialize_repo("tests/specs/bash_step/bash_step.yml")
    with repo_initializer.initialize() as r:
        dirs = os.listdir(r.working_dir)
        assert (
            len(
                set(dirs)
                & {"file1.txt", "file2.txt", "file3.txt", "file4.txt", "file5.txt"}
            )
            == 5
        )
        with open(os.path.join(r.working_dir, "file4.txt"), "r") as file:
            assert file.readlines() == ["Hello world\n"]



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class BranchStep(Step):
    branch_name: str

    step_type: StepType = field(init=False, default=StepType.BRANCH)

    def execute(self, repo: Repo) -> None:
        # TODO: Handle when attempting to create a branch when no commits exist
        branch = repo.create_head(self.branch_name)
        branch.checkout()

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "branch-name" not in step:
            raise ValueError('Missing "branch-name" field in branch step.')

        if step["branch-name"] is None or step["branch-name"].strip() == "":
            raise ValueError('Empty "branch-name" field in branch step.')

        return cls(
            name=name,
            description=description,
            id=id,
            branch_name=step["branch-name"],
        )



import pytest

from repo_smith.steps.remote_step import RemoteStep


def test_remote_step_parse_missing_remote_url():
    with pytest.raises(ValueError, match='Missing "remote-url" field in remote step.'):
        RemoteStep.parse("n", "d", "id", {})


def test_remote_step_parse_empty_remote_url():
    with pytest.raises(ValueError, match='Empty "remote-url" field in remote step.'):
        RemoteStep.parse("n", "d", "id", {"remote-url": ""})


def test_remote_step_parse_missing_remote_name():
    with pytest.raises(ValueError, match='Missing "remote-name" field in remote step.'):
        RemoteStep.parse("n", "d", "id", {"remote-url": "https://test.com"})


def test_remote_step_parse_empty_remote_name():
    with pytest.raises(ValueError, match='Empty "remote-name" field in remote step.'):
        RemoteStep.parse(
            "n", "d", "id", {"remote-url": "https://test.com", "remote-name": ""}
        )


def test_remote_step_parse():
    step = RemoteStep.parse(
        "n", "d", "id", {"remote-url": "https://test.com", "remote-name": "upstream"}
    )
    assert isinstance(step, RemoteStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.remote_url == "https://test.com"
    assert step.remote_name == "upstream"



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class RemoteStep(Step):
    remote_name: str
    remote_url: str

    step_type: StepType = field(init=False, default=StepType.REMOTE)

    def execute(self, repo: Repo) -> None:
        repo.create_remote(self.remote_name, self.remote_url)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "remote-url" not in step:
            raise ValueError('Missing "remote-url" field in remote step.')

        if step["remote-url"] is None or step["remote-url"].strip() == "":
            raise ValueError('Empty "remote-url" field in remote step.')

        if "remote-name" not in step:
            raise ValueError('Missing "remote-name" field in remote step.')

        if step["remote-name"] is None or step["remote-name"].strip() == "":
            raise ValueError('Empty "remote-name" field in remote step.')

        return cls(
            name=name,
            description=description,
            id=id,
            remote_name=step["remote-name"],
            remote_url=step["remote-url"],
        )



import os

from git import Repo
import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_add_step_missing_files():
    with pytest.raises(ValueError, match='Missing "files" field in add step.'):
        initialize_repo("tests/specs/add_step/add_step_missing_files.yml")


def test_add_step_empty_files():
    with pytest.raises(ValueError, match='Empty "files" list in add step.'):
        initialize_repo("tests/specs/add_step/add_step_empty_files.yml")


def test_add_step():
    def pre_hook(r: Repo) -> None:
        filename = os.path.join(r.working_dir, "file.txt")
        assert os.path.isfile(filename)

    repo_initializer = initialize_repo("tests/specs/add_step/add_step.yml")
    repo_initializer.add_pre_hook("add", pre_hook)

    with repo_initializer.initialize() as r:
        assert len(r.index.entries) == 1
        assert ("file.txt", 0) in r.index.entries



import pytest
from git import Repo
from src.repo_smith.initialize_repo import initialize_repo

# TODO: Test to make sure that the YAML parsing is accurate so we avoid individual
# integration test for every corner case covered in unit tests


def test_initialize_repo_missing_spec_path():
    with pytest.raises(ValueError, match="Invalid spec_path provided, not found."):
        initialize_repo("tests/specs/invalid_spec_path_does_not_exist.yml")


def test_initialize_repo_incomplete_spec_file():
    with pytest.raises(ValueError, match="Incomplete spec file."):
        initialize_repo("tests/specs/incomplete_spec_file.yml")


def test_initialize_repo_duplicate_ids():
    with pytest.raises(
        ValueError,
        match="ID commit is duplicated from a previous step. All IDs should be unique.",
    ):
        initialize_repo("tests/specs/duplicate_ids.yml")


def test_initialize_repo_duplicate_tags():
    with pytest.raises(
        ValueError,
        match="Tag tag is already in use by a previous step. All tag names should be unique.",
    ):
        initialize_repo("tests/specs/duplicate_tags.yml")


def test_initialize_repo_invalid_pre_hook():
    with pytest.raises(Exception):
        repo_initializer = initialize_repo("tests/specs/basic_spec.yml")
        repo_initializer.add_pre_hook("hello-world", lambda _: None)


def test_initialize_repo_invalid_post_hook():
    with pytest.raises(Exception):
        repo_initializer = initialize_repo("tests/specs/basic_spec.yml")
        repo_initializer.add_post_hook("hello-world", lambda _: None)


def test_initialize_repo_pre_hook():
    def initial_commit_pre_hook(_: Repo):
        assert True

    repo_initializer = initialize_repo("tests/specs/basic_spec.yml")
    repo_initializer.add_pre_hook("initial-commit", initial_commit_pre_hook)
    with repo_initializer.initialize() as r:
        assert r.commit("start-tag") is not None


def test_initialize_repo_post_hook():
    def initial_commit_post_hook(_: Repo):
        assert True

    repo_initializer = initialize_repo("tests/specs/basic_spec.yml")
    repo_initializer.add_post_hook("initial-commit", initial_commit_post_hook)
    with repo_initializer.initialize():
        pass


def test_initialize_repo_basic_spec():
    initialize_repo("tests/specs/basic_spec.yml")


def test_initialize_repo_hooks():
    initialize_repo("tests/specs/hooks.yml")



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class TagOptions(TypedDict, total=False):
    message: str
    force: bool
    delete: bool
    annotate: bool


TAG_SPEC = (
    CommandSpec()
    .opt("message", "-m")
    .flag("force", "-f", default=False)
    .flag("delete", "-d", default=False)
    .flag("annotate", "-a")
)



import pytest
from src.repo_smith.initialize_repo import initialize_repo


def test_tag_step_missing_tag_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/tag_step/tag_step_missing_tag_name.yml")


def test_tag_step_empty_tag_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/tag_step/tag_step_empty_tag_name.yml")


def test_tag_step_invalid_tag_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/tag_step/tag_step_invalid_tag_name.yml")


def test_tag_step():
    repo_initializer = initialize_repo("tests/specs/tag_step/tag_step.yml")
    with repo_initializer.initialize() as r:
        assert len(r.tags) == 2
        assert r.tags[0].tag is not None
        assert r.tags[0].tag.tag == "with-description"
        assert r.tags[0].tag.message == "Hello world!"
        assert r.tags[1].tag is None
        assert str(r.tags[1]) == "without-description"



import abc
from dataclasses import dataclass
from typing import Any, Callable, Dict, List, Mapping, Self, TypedDict


@dataclass(frozen=True)
class Opt:
    flag: str
    takes_value: bool = True
    transform: Callable[[Any], str] = str
    default: Any = None


def build_args(options: Mapping[str, Any], specs: Mapping[str, Opt]) -> List[str]:
    args = []
    for key, value in options.items():
        if value is None:
            continue

        spec = specs.get(key)
        if spec is None:
            raise ValueError(f"Unsupported option: {key}")

        if spec.takes_value:
            args.append(spec.flag)
            args.append(spec.transform(value))
        elif value:
            args.append(spec.flag)
    return args


class CommandSpec:
    def __init__(self) -> None:
        self._specs: Dict[str, Opt] = {}
        self._defaults: Dict[str, Any] = {}

    def opt(
        self,
        name: str,
        flag: str,
        *,
        default: Any = None,
        transform: Callable[[Any], str] = str,
    ) -> Self:
        self._specs[name] = Opt(flag, True, transform, default)
        if default is not None:
            self._defaults[name] = default
        return self

    def flag(
        self,
        name: str,
        flag: str,
        *,
        default: Any = None,
    ) -> Self:
        self._specs[name] = Opt(flag, False, str, default)
        if default is not None:
            self._defaults[name] = default
        return self

    def build(self, options: Mapping[str, Any]) -> List[str]:
        merged = dict(self._defaults)
        merged.update(options)
        return build_args(merged, self._specs)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step_type import StepType


@dataclass
class Step(ABC):
    name: Optional[str]
    step_type: StepType
    description: Optional[str]
    id: Optional[str]

    @abstractmethod
    def execute(self, repo: Repo) -> None:
        pass

    @classmethod
    @abstractmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        pass



from unittest.mock import patch

import pytest
from repo_smith.steps.add_step import AddStep
from repo_smith.steps.bash_step import BashStep
from repo_smith.steps.branch_delete_step import BranchDeleteStep
from repo_smith.steps.branch_rename_step import BranchRenameStep
from repo_smith.steps.branch_step import BranchStep
from repo_smith.steps.checkout_step import CheckoutStep
from repo_smith.steps.commit_step import CommitStep
from repo_smith.steps.dispatcher import Dispatcher
from repo_smith.steps.fetch_step import FetchStep
from repo_smith.steps.file_step import (
    AppendFileStep,
    DeleteFileStep,
    EditFileStep,
    NewFileStep,
)
from repo_smith.steps.merge_step import MergeStep
from repo_smith.steps.remote_step import RemoteStep
from repo_smith.steps.step_type import StepType
from repo_smith.steps.tag_step import TagStep


def test_dispatch_calls_correct_step_parse():
    step_dict = {
        "type": "commit",
        "name": "my commit",
        "description": "desc",
        "id": "123",
    }

    # Patch CommitStep.parse so we don't run real logic
    with patch(
        "repo_smith.steps.commit_step.CommitStep.parse", return_value="parsed"
    ) as mock_parse:
        result = Dispatcher.dispatch(step_dict)

    # parse should have been called once with correct arguments
    mock_parse.assert_called_once_with("my commit", "desc", "123", step_dict)

    # The dispatcher should return whatever parse returned
    assert result == "parsed"


def test_dispatch_missing_type_raises():
    step_dict = {"name": "no type"}
    with pytest.raises(ValueError, match='Missing "type" field in step.'):
        Dispatcher.dispatch(step_dict)


STEP_TYPES_TO_CLASSES = {
    StepType.COMMIT: CommitStep,
    StepType.ADD: AddStep,
    StepType.TAG: TagStep,
    StepType.NEW_FILE: NewFileStep,
    StepType.EDIT_FILE: EditFileStep,
    StepType.DELETE_FILE: DeleteFileStep,
    StepType.APPEND_FILE: AppendFileStep,
    StepType.BASH: BashStep,
    StepType.BRANCH: BranchStep,
    StepType.BRANCH_RENAME: BranchRenameStep,
    StepType.BRANCH_DELETE: BranchDeleteStep,
    StepType.CHECKOUT: CheckoutStep,
    StepType.REMOTE: RemoteStep,
    StepType.MERGE: MergeStep,
    StepType.FETCH: FetchStep,
}


@pytest.mark.parametrize("step_type, step_path", STEP_TYPES_TO_CLASSES.items())
def test_get_type_returns_correct_class(step_type, step_path):
    # Uses mangled name: https://stackoverflow.com/questions/2064202/private-members-in-python
    assert Dispatcher._Dispatcher__get_type(step_type) is step_path  # type: ignore



import pytest
from repo_smith.steps.tag_step import TagStep


def test_tag_step_parse_missing_tag_name():
    with pytest.raises(ValueError, match='Missing "tag-name" field in tag step.'):
        TagStep.parse("n", "d", "id", {})


def test_tag_step_parse_empty_tag_name():
    with pytest.raises(ValueError, match='Empty "tag-name" field in tag step.'):
        TagStep.parse("n", "d", "id", {"tag-name": ""})


def test_tag_step_parse_invalid_tag_name():
    with pytest.raises(
        ValueError,
        match='Field "tag-name" can only contain alphanumeric characters, _, -, .',
    ):
        TagStep.parse("n", "d", "id", {"tag-name": "(open)"})


def test_tag_step_parse_missing_tag_message():
    step = TagStep.parse("n", "d", "id", {"tag-name": "start"})
    assert isinstance(step, TagStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.tag_name == "start"
    assert step.tag_message is None


def test_tag_step_parse_with_tag_message():
    step = TagStep.parse(
        "n", "d", "id", {"tag-name": "start", "tag-message": "this is a message"}
    )
    assert isinstance(step, TagStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.tag_name == "start"
    assert step.tag_message == "this is a message"



import os

import pytest
from git import Repo

from src.repo_smith.initialize_repo import initialize_repo


def test_append_file_step_missing_filename():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/append_file_step/append_file_missing_filename.yml")


def test_append_file_step_no_filename():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/append_file_step/append_file_no_filename.yml")


def test_append_file_step():
    def pre_hook(r: Repo) -> None:
        filepath = os.path.join(r.working_dir, "file.txt")
        assert os.path.isfile(filepath)
        expected = ["Hello world"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected)
            for actual, ex in zip(lines, expected):
                assert actual == ex

    def post_hook(r: Repo) -> None:
        filepath = os.path.join(r.working_dir, "file.txt")
        assert os.path.isfile(filepath)
        expected = ["Hello world", "This is a new line!"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected)
            for actual, ex in zip(lines, expected):
                assert actual == ex

    repo_initializer = initialize_repo("tests/specs/append_file_step/append_file.yml")
    repo_initializer.add_pre_hook("append-file", pre_hook)
    repo_initializer.add_post_hook("append-file", post_hook)
    with repo_initializer.initialize():
        pass



from git import Repo
import pytest

from src.repo_smith.initialize_repo import initialize_repo


def test_checkout_step_missing_branch_name():
    with pytest.raises(Exception):
        initialize_repo(
            "tests/specs/checkout_step/checkout_step_missing_branch_name.yml"
        )


def test_checkout_step_empty_branch_name():
    with pytest.raises(Exception):
        initialize_repo("tests/specs/checkout_step/checkout_step_empty_branch_name.yml")


def test_checkout_step_missing_branch():
    with pytest.raises(Exception):
        repo_initialier = initialize_repo(
            "tests/specs/checkout_step/checkout_step_missing_branch.yml"
        )
        with repo_initialier.initialize():
            pass


def test_checkout_step():
    def first_hook(r: Repo) -> None:
        assert r.active_branch.name == "main"

    def second_hook(r: Repo) -> None:
        assert r.active_branch.name == "test"

    repo_initializer = initialize_repo("tests/specs/checkout_step/checkout_step.yml")
    repo_initializer.add_post_hook("first", first_hook)
    repo_initializer.add_post_hook("second", second_hook)
    with repo_initializer.initialize() as r:
        assert len(r.branches) == 2
        assert "test" in r.heads



from repo_smith.initialize_repo import initialize_repo

# TODO: more corner case testing


def test_merge_step_squash():
    ir = initialize_repo("tests/specs/merge_step/merge_step_squash.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits())
        commit_messages = [c.message.strip() for c in commits][::-1]
        assert commit_messages == ["Before", "Squash merge branch 'incoming'", "After"]


def test_merge_step_no_fast_forward():
    ir = initialize_repo("tests/specs/merge_step/merge_step_no_fast_forward.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits())
        commit_messages = [c.message.strip() for c in commits][::-1]
        assert "Merge branch 'incoming'" in commit_messages
        assert len(commits[1].parents) == 2


def test_merge_step_with_fast_forward():
    ir = initialize_repo("tests/specs/merge_step/merge_step_with_fast_forward.yml")
    with ir.initialize() as r:
        commits = list(r.iter_commits())
        commit_messages = [c.message.strip() for c in commits][::-1]
        assert "Merge branch 'incoming'" not in commit_messages
        assert not any([len(c.parents) > 1 for c in commits])



import pytest
from repo_smith.steps.file_step import (
    AppendFileStep,
    DeleteFileStep,
    EditFileStep,
    FileStep,
    NewFileStep,
)


def test_file_step_get_details_missing_filename():
    with pytest.raises(ValueError, match='Missing "filename" field in file step.'):
        FileStep.get_details({})


def test_file_step_get_details_empty_filename():
    with pytest.raises(ValueError, match='Empty "filename" field in file step.'):
        FileStep.get_details({"filename": ""})


def test_file_step_get_details_missing_contents():
    filename, contents = FileStep.get_details({"filename": "hello.txt"})
    assert filename == "hello.txt"
    assert contents == ""


def test_file_step_get_details():
    filename, contents = FileStep.get_details(
        {"filename": "hello.txt", "contents": "Hello world"}
    )
    assert filename == "hello.txt"
    assert contents == "Hello world"


FILE_STEP_CLASSES = [NewFileStep, EditFileStep, AppendFileStep, DeleteFileStep]


@pytest.mark.parametrize("step_class", FILE_STEP_CLASSES)
def test_new_file_step_parse(step_class):
    step = step_class.parse(
        "n", "d", "id", {"filename": "hello.txt", "contents": "Hello world"}
    )
    assert isinstance(step, step_class)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.filename == "hello.txt"
    assert step.contents == "Hello world"



from typing import List, Optional, Union, Unpack

from repo_smith.helpers.git_helper.add_options import ADD_SPEC, AddOptions
from repo_smith.helpers.git_helper.checkout_options import (
    CHECKOUT_SPEC,
    CheckoutOptions,
)
from repo_smith.helpers.git_helper.commit_options import COMMIT_SPEC, CommitOptions
from repo_smith.helpers.git_helper.remote_options import (
    REMOTE_ADD_SPEC,
    RemoteAddOptions,
)
from repo_smith.helpers.git_helper.restore_options import RESTORE_SPEC, RestoreOptions
from repo_smith.helpers.git_helper.tag_options import TAG_SPEC, TagOptions
from repo_smith.helpers.helper import Helper


class GitHelper(Helper):
    def __init__(self, verbose: bool) -> None:
        super().__init__(verbose)

    def tag(
        self,
        tag_name: str,
        object_id: Optional[str] = None,
        **options: Unpack[TagOptions],
    ) -> None:
        """Calls the underlying git-tag command with the given support options.

        More information about the git-tag command can be found `here <https://git-scm.com/docs/git-tag>`__.
        """
        args = ["git", "tag", tag_name]
        if object_id is not None:
            args.append(object_id)
        args.extend(TAG_SPEC.build(options))
        self.run(args)

    def add(
        self,
        files: Optional[Union[str, List[str]]] = None,
        **options: Unpack[AddOptions],
    ) -> None:
        """Calls the underlying git-add command with the given support options.

        More information about the git-add command can be found `here <https://git-scm.com/docs/git-add>`__.
        """
        if files is None:
            files = []
        elif isinstance(files, str):
            files = [files]
        args = ["git", "add", *files] + ADD_SPEC.build(options)
        self.run(args)

    # TODO: Create a class just for the pathspec format
    def commit(
        self,
        pathspec: Optional[str] = None,
        **options: Unpack[CommitOptions],
    ) -> None:
        """Calls the underlying git-commit command with the given support options.

        More information about the git-commit command can be found `here <https://git-scm.com/docs/git-commit>`__.
        """
        trailing = [] if pathspec is None else [pathspec]
        args = ["git", "commit"] + COMMIT_SPEC.build(options) + trailing
        self.run(args)

    def remote_add(
        self,
        name: str,
        url: str,
        **options: Unpack[RemoteAddOptions],
    ) -> None:
        """Calls the underlying git-remote add command with the given support options.

        More information about the git-remote add command can be found `here <https://git-scm.com/docs/git-remote>`__.
        """
        args = ["git", "remote", "add"] + REMOTE_ADD_SPEC.build(options) + [name, url]
        self.run(args)

    def remote_rename(self, old: str, new: str) -> None:
        """Calls the underlying git-remote rename command with the given support options.

        More information about the git-remote rename command can be found `here <https://git-scm.com/docs/git-remote>`__.
        """
        args = ["git", "remote", "rename", old, new]
        self.run(args)

    def remote_remove(self, name: str) -> None:
        """Calls the underlying git-remote remove command with the given support options.

        More information about the git-remote remove command can be found `here <https://git-scm.com/docs/git-remote>`__.
        """
        args = ["git", "remote", "remove", name]
        self.run(args)

    def checkout(
        self,
        branch_name: Optional[str] = None,
        start_point: Optional[str] = None,
        paths: Optional[Union[str, List[str]]] = None,
        **options: Unpack[CheckoutOptions],
    ) -> None:
        """Calls the underlying git-checkout command with the given support options.

        More information about the git-checkout command can be found `here <https://git-scm.com/docs/git-checkout>`__.
        """
        # git-checkout prioritizes the branch first, so if the branch is provided, we use it first
        if paths is not None and len(paths) > 0:
            if options.get("branch", False):
                # The alternative is to just ignore this field if set
                raise ValueError("Cannot use '-b' when specifying paths.")

        args = ["git", "checkout"] + CHECKOUT_SPEC.build(options)
        if branch_name is not None:
            args.append(branch_name)

        if start_point is not None:
            # So we assume checking out by branch, otherwise we look for files
            args.append(start_point)
        elif paths is not None:
            args.append("--")
            paths = [paths] if isinstance(paths, str) else paths
            args.extend(paths)

        self.run(args)

    def restore(
        self,
        pathspec: Optional[str] = None,
        **options: Unpack[RestoreOptions],
    ) -> None:
        """Calls the underlying git-restore command with the given support options.

        More information about the git-restore command can be found `here <https://git-scm.com/docs/git-restore>`__.
        """
        if (
            options.get("ours", False)
            or options.get("theirs", False)
            or options.get("merge", False)
            or options.get("conflict") is not None
        ) and options.get("source") is not None:
            raise ValueError(
                "Cannot use --ours, --theirs, --merge, or --conflict with --source."
            )

        if (
            options.get("ours", False)
            or options.get("theirs", False)
            or options.get("merge", False)
            or options.get("conflict") is not None
        ) and options.get("ignore_unmerged") is not None:
            raise ValueError(
                "Cannot use --ours, --theirs, --merge, or --conflict with --ignore-unmerged."
            )

        trailing = [] if pathspec is None else [pathspec]
        args = ["git", "restore"] + RESTORE_SPEC.build(options) + trailing
        self.run(args)



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import BadName, Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class CheckoutStep(Step):
    branch_name: Optional[str]
    commit_hash: Optional[str]

    step_type: StepType = field(init=False, default=StepType.CHECKOUT)

    def execute(self, repo: Repo) -> None:
        if self.branch_name is not None:
            if self.branch_name not in repo.heads:
                raise ValueError("Invalid branch name")
            else:
                repo.heads[self.branch_name].checkout()

        if self.commit_hash:
            try:
                commit = repo.commit(self.commit_hash)
                repo.git.checkout(commit)
            except (ValueError, BadName):
                raise ValueError("Commit not found")

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if step.get("branch-name") is None and step.get("commit-hash") is None:
            raise ValueError(
                'Provide either "branch-name" or "commit-hash" in checkout step.'
            )

        if step.get("branch-name") is not None and step.get("commit-hash") is not None:
            raise ValueError(
                'Provide either "branch-name" or "commit-hash", not both, in checkout step.'
            )

        if step.get("branch-name") is not None and step["branch-name"].strip() == "":
            raise ValueError('Empty "branch-name" field in checkout step.')

        if step.get("commit-hash") is not None and step["commit-hash"].strip() == "":
            raise ValueError('Empty "commit-hash" field in checkout step.')

        return cls(
            name=name,
            description=description,
            id=id,
            branch_name=step.get("branch-name"),
            commit_hash=step.get("commit-hash"),
        )



import pytest

from repo_smith.steps.fetch_step import FetchStep


def test_fetch_step_parse_missing_remote_name():
    with pytest.raises(ValueError, match='Missing "remote-name" field in fetch step.'):
        FetchStep.parse("n", "d", "id", {})


def test_fetch_step_parse_empty_remote_name():
    with pytest.raises(ValueError, match='Empty "remote-name" field in fetch step.'):
        FetchStep.parse("n", "d", "id", {"remote-name": ""})


def test_commit_step_parse_with_empty():
    step = FetchStep.parse("n", "d", "id", {"remote-name": "test"})
    assert isinstance(step, FetchStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.remote_name == "test"



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class BranchRenameStep(Step):
    original_branch_name: str
    target_branch_name: str

    step_type: StepType = field(init=False, default=StepType.BRANCH)

    def execute(self, repo: Repo) -> None:
        if self.original_branch_name not in repo.heads:
            raise ValueError(
                '"branch-name" field provided does not correspond to any existing branches in branch-rename step.'
            )
        if self.target_branch_name in repo.heads:
            raise ValueError(
                '"new-name" field provided corresponds to an existing branch already in branch-rename step.'
            )
        branch = repo.heads[self.original_branch_name]
        branch.rename(self.target_branch_name)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "branch-name" not in step:
            raise ValueError('Missing "branch-name" field in branch-rename step.')

        if step["branch-name"] is None or step["branch-name"].strip() == "":
            raise ValueError('Empty "branch-name" field in branch-rename step.')

        if "new-name" not in step:
            raise ValueError('Missing "new-name" field in branch-rename step.')

        if step["new-name"] is None or step["new-name"].strip() == "":
            raise ValueError('Empty "new-name" field in branch-rename step.')

        return cls(
            name=name,
            description=description,
            id=id,
            original_branch_name=step["branch-name"],
            target_branch_name=step["new-name"],
        )



from enum import Enum


class StepType(Enum):
    COMMIT = "commit"
    ADD = "add"
    TAG = "tag"
    NEW_FILE = "new-file"
    EDIT_FILE = "edit-file"
    DELETE_FILE = "delete-file"
    APPEND_FILE = "append-file"
    BASH = "bash"
    BRANCH = "branch"
    BRANCH_RENAME = "branch-rename"
    BRANCH_DELETE = "branch-delete"
    CHECKOUT = "checkout"
    REMOTE = "remote"
    MERGE = "merge"
    FETCH = "fetch"

    @staticmethod
    def from_value(value: str) -> "StepType":
        match value:
            case "commit":
                return StepType.COMMIT
            case "add":
                return StepType.ADD
            case "tag":
                return StepType.TAG
            case "new-file":
                return StepType.NEW_FILE
            case "edit-file":
                return StepType.EDIT_FILE
            case "delete-file":
                return StepType.DELETE_FILE
            case "append-file":
                return StepType.APPEND_FILE
            case "bash":
                return StepType.BASH
            case "branch":
                return StepType.BRANCH
            case "branch-rename":
                return StepType.BRANCH_RENAME
            case "branch-delete":
                return StepType.BRANCH_DELETE
            case "checkout":
                return StepType.CHECKOUT
            case "remote":
                return StepType.REMOTE
            case "merge":
                return StepType.MERGE
            case "fetch":
                return StepType.FETCH
            case _:
                raise ValueError(f"Invalid value {value} given. Not supported.")



from typing import TypedDict

from repo_smith.helpers.command_spec import CommandSpec


class AddOptions(TypedDict, total=False):
    force: bool
    update: bool
    all: bool


ADD_SPEC = (
    CommandSpec()
    .flag("force", "-f", default=False)
    .flag("update", "-u", default=False)
    .flag("all", "-a", default=False)
)



import tempfile
from contextlib import contextmanager
from typing import Iterator, TypedDict, Unpack

from git.repo import Repo

from repo_smith.helpers.files_helper import FilesHelper
from repo_smith.helpers.git_helper.git_helper import GitHelper


class RepoSmith:
    def __init__(self, repo: Repo, verbose: bool) -> None:
        self.verbose = verbose
        self.repo = repo
        self.files = FilesHelper(verbose)
        self.git = GitHelper(verbose)


class CreateRepoOptions(TypedDict, total=False):
    clone_from: str
    existing_path: str


@contextmanager
def create_repo_smith(
    verbose: bool, **options: Unpack[CreateRepoOptions]
) -> Iterator[RepoSmith]:
    """Creates a RepoSmith instance over a given repository."""
    clone_from = options.get("clone_from")
    existing_path = options.get("existing_path")

    dir = tempfile.mkdtemp() if existing_path is None else existing_path

    if clone_from:
        repo = Repo.clone_from(clone_from, dir)
    else:
        repo = Repo.init(dir, initial_branch="main")

    yield RepoSmith(repo, verbose)



import os

import pytest
from git import Repo
from repo_smith.initialize_repo import initialize_repo
from tests.fixtures.git_fixtures import REMOTE_REPO_PATH, remote_repo


def test_fetch_step_remote_valid(remote_repo: Repo):
    (REMOTE_REPO_PATH / "dummy.txt").write_text("initial")
    remote_repo.index.add(["dummy.txt"])
    remote_repo.index.commit("initial commit")

    remote_repo_commit_hexsha = remote_repo.commit("main").hexsha
    ir = initialize_repo("tests/specs/fetch_step/fetch_step_remote_valid.yml")
    with ir.initialize() as r:
        latest_commit_hexsha = r.commit("origin/main").hexsha
        assert latest_commit_hexsha == remote_repo_commit_hexsha


def test_fetch_step_missing_remote(remote_repo: Repo):
    (REMOTE_REPO_PATH / "dummy.txt").write_text("initial")
    remote_repo.index.add(["dummy.txt"])
    remote_repo.index.commit("initial commit")

    ir = initialize_repo("tests/specs/fetch_step/fetch_step_missing_remote.yml")
    with pytest.raises(ValueError, match="Missing remote 'upstream' in fetch step."):
        with ir.initialize() as _:
            pass



import os

import pytest
from git import Repo

from src.repo_smith.initialize_repo import initialize_repo


def test_edit_file_step_unknown_file() -> None:
    with pytest.raises(Exception):
        repo_initializer = initialize_repo(
            "tests/specs/edit_file_step/edit_file_missing.yml"
        )
        with repo_initializer.initialize():
            pass


def test_edit_file_step() -> None:
    def add_hook(r: Repo) -> None:
        dir_list = os.listdir(r.working_dir)
        assert "filea.txt" in dir_list
        filepath = os.path.join(r.working_dir, "filea.txt")
        expected_file_contents = ["Original text"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected_file_contents)
            for actual, expected in zip(lines, expected_file_contents):
                assert actual == expected

    def edit_hook(r: Repo) -> None:
        dir_list = os.listdir(r.working_dir)
        assert "filea.txt" in dir_list
        filepath = os.path.join(r.working_dir, "filea.txt")
        expected_file_contents = ["Edited text"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected_file_contents)
            for actual, expected in zip(lines, expected_file_contents):
                assert actual == expected

    repo_initializer = initialize_repo("tests/specs/edit_file_step/edit_file.yml")
    repo_initializer.add_post_hook("add", add_hook)
    repo_initializer.add_post_hook("edit", edit_hook)
    with repo_initializer.initialize():
        pass



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class CommitStep(Step):
    empty: bool
    message: str

    step_type: StepType = field(init=False, default=StepType.COMMIT)

    def execute(self, repo: Repo) -> None:
        if self.empty:
            repo.git.commit("-m", self.message, "--allow-empty")
        else:
            repo.index.commit(message=self.message)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "message" not in step:
            raise ValueError('Missing "message" field in commit step.')

        if step["message"] is None or step["message"].strip() == "":
            raise ValueError('Empty "message" field in commit step.')

        return cls(
            name=name,
            description=description,
            id=id,
            empty=step.get("empty", False),
            message=step["message"],
        )



from dataclasses import dataclass, field
from typing import Any, List, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class AddStep(Step):
    files: List[str]

    step_type: StepType = field(init=False, default=StepType.ADD)

    def execute(self, repo: Repo) -> None:
        repo.index.add(self.files)

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "files" not in step:
            raise ValueError('Missing "files" field in add step.')

        if step["files"] is None or step["files"] == []:
            raise ValueError('Empty "files" list in add step.')

        return cls(
            name=name,
            description=description,
            id=id,
            files=step["files"],
        )



import os
import subprocess
from dataclasses import dataclass
from subprocess import CompletedProcess
from sys import exit
from typing import Dict, List


@dataclass
class CommandResult:
    result: CompletedProcess[str]

    def is_success(self) -> bool:
        return self.result.returncode == 0

    @property
    def stdout(self) -> str:
        return self.result.stdout.strip()

    @property
    def returncode(self) -> int:
        return self.result.returncode


def run(
    command: List[str],
    verbose: bool,
    env: Dict[str, str] = {},
    exit_on_error: bool = False,
) -> CommandResult:
    """Runs the given command, logging the output if verbose is True."""
    try:
        result = subprocess.run(
            command,
            capture_output=True,
            text=True,
            env=dict(os.environ, **env),
            encoding="utf-8",
        )
    except FileNotFoundError:
        if exit_on_error:
            exit(1)
        error_msg = f"Command not found: {command[0]}"
        result = CompletedProcess(command, returncode=127, stdout="", stderr=error_msg)
    except PermissionError:
        if exit_on_error:
            exit(1)
        error_msg = f"Permission denied: {command[0]}"
        result = CompletedProcess(command, returncode=126, stdout="", stderr=error_msg)
    except OSError as e:
        if exit_on_error:
            exit(1)
        error_msg = f"OS error when running command {command}: {e}"
        result = CompletedProcess(command, returncode=1, stdout="", stderr=error_msg)

    if verbose:
        if result.returncode == 0:
            print("\t" + result.stdout)
        else:
            print("\t" + result.stderr)

    return CommandResult(result=result)



import pytest
from repo_smith.initialize_repo import initialize_repo


def test_branch_rename_step_branch_exists():
    repo_initializer = initialize_repo(
        "tests/specs/branch_rename_step/branch_rename_step_branch_exists.yml"
    )
    with repo_initializer.initialize() as r:
        assert {r.name for r in r.refs} == {"main", "primary"}


def test_branch_rename_step_branch_does_not_exist():
    repo_initializer = initialize_repo(
        "tests/specs/branch_rename_step/branch_rename_step_branch_does_not_exist.yml"
    )
    with pytest.raises(
        ValueError,
        match='"branch-name" field provided does not correspond to any existing branches in branch-rename step.',
    ):
        with repo_initializer.initialize() as _:
            pass


def test_branch_rename_step_branch_already_existed():
    repo_initializer = initialize_repo(
        "tests/specs/branch_rename_step/branch_rename_step_branch_already_existed.yml"
    )
    with pytest.raises(
        ValueError,
        match='"new-name" field provided corresponds to an existing branch already in branch-rename step.',
    ):
        with repo_initializer.initialize() as _:
            pass



from dataclasses import dataclass


@dataclass
class CloneFrom:
    """Indicates that the first step of the initialization should be to clone the
    indicated repository and then apply commits to it.
    """

    repo_url: str



import pytest
from repo_smith.initialize_repo import initialize_repo


def test_branch_delete_step_branch_exists():
    repo_initializer = initialize_repo(
        "tests/specs/branch_delete_step/branch_delete_step_branch_exists.yml"
    )
    with repo_initializer.initialize() as r:
        assert [r.name for r in r.refs] == ["main"]


def test_branch_delete_step_branch_does_not_exist():
    repo_initializer = initialize_repo(
        "tests/specs/branch_delete_step/branch_delete_step_branch_does_not_exist.yml"
    )

    with pytest.raises(
        ValueError,
        match='"branch-name" field provided does not correspond to any existing branches in branch-delete step.',
    ):
        with repo_initializer.initialize() as _:
            pass



import pytest

from repo_smith.steps.checkout_step import CheckoutStep


def test_checkout_step_parse_missing_branch_name_and_commit_hash():
    with pytest.raises(
        ValueError,
        match='Provide either "branch-name" or "commit-hash" in checkout step.',
    ):
        CheckoutStep.parse("n", "d", "id", {})


def test_checkout_step_parse_both_branch_name_and_commit_hash():
    with pytest.raises(
        ValueError,
        match='Provide either "branch-name" or "commit-hash", not both, in checkout step.',
    ):
        CheckoutStep.parse(
            "n", "d", "id", {"branch-name": "test", "commit-hash": "abc123"}
        )


def test_checkout_step_parse_empty_branch_name():
    with pytest.raises(ValueError, match='Empty "branch-name" field in checkout step.'):
        CheckoutStep.parse("n", "d", "id", {"branch-name": ""})


def test_checkout_step_parse_empty_commit_hash():
    with pytest.raises(ValueError, match='Empty "commit-hash" field in checkout step.'):
        CheckoutStep.parse("n", "d", "id", {"commit-hash": ""})


def test_checkout_step_parse_with_branch_name():
    step = CheckoutStep.parse("n", "d", "id", {"branch-name": "test"})
    assert isinstance(step, CheckoutStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.branch_name == "test"
    assert step.commit_hash is None


def test_checkout_step_parse_with_commit_hash():
    step = CheckoutStep.parse("n", "d", "id", {"commit-hash": "abc123"})
    assert isinstance(step, CheckoutStep)
    assert step.name == "n"
    assert step.description == "d"
    assert step.id == "id"
    assert step.branch_name is None
    assert step.commit_hash == "abc123"



import os

import pytest
from git import Repo

from src.repo_smith.initialize_repo import initialize_repo


def test_new_file_step_missing_filename() -> None:
    with pytest.raises(Exception):
        initialize_repo("tests/specs/new_file_step/new_file_missing_filename.yml")


def test_new_file_step_empty_filename() -> None:
    with pytest.raises(Exception):
        initialize_repo("tests/specs/new_file_step/new_file_empty_filename.yml")


def test_new_file_step_empty_contents() -> None:
    repo_initializer = initialize_repo(
        "tests/specs/new_file_step/new_file_empty_contents.yml"
    )
    with repo_initializer.initialize() as r:
        filea = os.path.join(r.working_dir, "filea.txt")
        fileb = os.path.join(r.working_dir, "fileb.txt")
        assert os.path.isfile(filea)
        assert os.path.isfile(fileb)
        assert os.path.getsize(filea) == 0
        assert os.path.getsize(fileb) == 0


def test_new_file_step() -> None:
    def validate_filea_hook(r: Repo) -> None:
        dir_list = os.listdir(r.working_dir)
        assert "filea.txt" in dir_list
        filepath = os.path.join(r.working_dir, "filea.txt")
        expected_file_contents = ["Hello world!", "", "This is a file"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected_file_contents)
            for actual, expected in zip(lines, expected_file_contents):
                assert actual == expected

    def validate_nested_file_hook(r: Repo) -> None:
        dir_list = os.listdir(r.working_dir)
        assert "nested" in dir_list
        filepath = os.path.join(r.working_dir, "nested/a/b/c/filed.txt")
        assert os.path.isfile(filepath)
        expected_file_contents = ["This is a nested file"]
        with open(filepath, "r") as f:
            lines = [line.strip() for line in f.readlines()]
            assert len(lines) == len(expected_file_contents)
            for actual, expected in zip(lines, expected_file_contents):
                assert actual == expected

    repo_initializer = initialize_repo("tests/specs/new_file_step/new_file.yml")
    repo_initializer.add_post_hook("filea", validate_filea_hook)
    repo_initializer.add_post_hook("nested_file", validate_nested_file_hook)
    with repo_initializer.initialize():
        pass



from dataclasses import dataclass, field
from typing import Any, Optional, Self, Type

from git import Repo
from repo_smith.steps.step import Step
from repo_smith.steps.step_type import StepType


@dataclass
class FetchStep(Step):
    remote_name: str

    step_type: StepType = field(init=False, default=StepType.FETCH)

    def execute(self, repo: Repo) -> None:
        try:
            remote = repo.remote(self.remote_name)
        except Exception:
            raise ValueError(f"Missing remote '{self.remote_name}' in fetch step.")

        remote.fetch()

    @classmethod
    def parse(
        cls: Type[Self],
        name: Optional[str],
        description: Optional[str],
        id: Optional[str],
        step: Any,
    ) -> Self:
        if "remote-name" not in step:
            raise ValueError('Missing "remote-name" field in fetch step.')

        if step["remote-name"] is None or step["remote-name"].strip() == "":
            raise ValueError('Empty "remote-name" field in fetch step.')

        return cls(
            name=name,
            description=description,
            id=id,
            remote_name=step["remote-name"],
        )



import pytest

from repo_smith.steps.bash_step import BashStep


def test_bash_step_parse_missing_runs():
    with pytest.raises(ValueError, match='Missing "runs" field in bash step.'):
        BashStep.parse("a", "d", "id", {})


def test_bash_step_parse_empty_runs():
    with pytest.raises(ValueError, match='Empty "runs" field in bash step.'):
        BashStep.parse("a", "d", "id", {"runs": ""})


def test_bash_step_parse():
    step = BashStep.parse("a", "d", "id", {"runs": "ls"})
    assert isinstance(step, BashStep)
    assert step.name == "a"
    assert step.description == "d"
    assert step.id == "id"
    assert step.body == "ls"
